import requests
from concurrent.futures import ThreadPoolExecutor,ProcessPoolExecutor,wait,ALL_COMPLETED
from tqdm import tqdm
import os
from loguru import logger
import tempfile
import random
import string


class MClient():
    '''
    多线程/进程下载客户端
    '''
    def __init__(self,url_list=[]):
        self.url_list = url_list

    def __download_file(self,url,save_path,timeout=10) -> None:
        """
        下载ts文件
        :param url: ts文件的URL
        :param save_path: 保存的路径
        """
        # 发送stream请求
        response = requests.get(url, stream=True,verify=False,timeout=timeout)
        
        response.raise_for_status()
        # 获取文件大小
        total_size = int(response.headers.get("Content-Length", 0))    
        # 显示下载进度
       
        progress = tqdm(response.iter_content(1024), f"Downloading {os.path.basename(save_path)}", total=total_size, unit="B", unit_scale=True, unit_divisor=1024)
       
       
        try:
            with open(save_path, "wb") as f:
                for data in progress.iterable:
                    # 写入文件
                    f.write(data)
                    # 更新进度条
                    progress.update(len(data))
        except requests.exceptions.RequestException:
            raise Exception(f"Download {save_path} failed")

    def _confirm_download_file(self,url,base_dir=".",timeout=10,retry_count=3):
        '''
        拥有尝试机制,下载文件
        '''
        c = 0
        save_path = f"{base_dir}/{os.path.basename(url)}"
        while c < retry_count:
            c += 1
            try:
                self.__download_file(url,save_path,timeout)
                break
            except Exception as e:
                logger.error(f"retry [{c}]...{url}\n{e}")

    def __init_output(self,save_path):
        '''
        初始化视频输出目录，
        没指定位置，默认同级out目录default_*****
        '''
        if not save_path:
            save_path = f"default_{''.join(random.choices(string.ascii_letters,k=5))}.mp4"
        out_dir = os.path.dirname(save_path)
        if out_dir and not os.path.exists(out_dir):
            os.makedirs(out_dir)
        return save_path

    def __download_ts(self,url,pbar,ts_save_path,retry=3):
        '''
        ts 文件下载
        '''
        c = 0
        while c < retry:
            c += 1
            try:
                with open(ts_save_path,"wb") as f:
                    f.write(requests.get(url,verify=False,timeout=20).content)
                    pbar.update()
                break
            except Exception as e:
                logger.error(f"retry [{c}]...{url}\n{e}")

    
    def download_vedio_info(self,save_path=None,workers=1,thread=True):
        '''
        单进度条显示视频下载进度
        '''
        with tempfile.TemporaryDirectory() as ts_temp_dir:
            save_path = self.__init_output(save_path)
            logger.debug(f"临时文件夹: {ts_temp_dir}")
            #多线程或进程下载，属性进度条
            total = len(self.url_list)
            with tqdm(desc=f"Downloading {os.path.basename(save_path)}", total=total, unit="个", unit_scale=True) as pbar:
                if thread:
                    executor = ThreadPoolExecutor(max_workers=workers)
                else:
                    executor = ProcessPoolExecutor(max_workers=workers)
                tasks = {executor.submit(self.__download_ts,url,pbar,f"{ts_temp_dir}/{os.path.basename(url)}"):f"{ts_temp_dir}/{os.path.basename(url)}" for url in self.url_list}
                wait(tasks,return_when=ALL_COMPLETED)
                logger.debug(f"ts 文件下载完成")
            # 合并ts文件
            
            with open(save_path,"wb+") as ww:
                for ts in tasks.values():
                    if os.path.exists(ts):
                        with open(ts,"rb") as rr:
                            ww.write(rr.read())
            logger.success(f"{save_path} 下载完成")
            return save_path

    def download_vedio(self,save_path=None,workers=1,thread=True,debug=False):
        if debug:
            return self.download_vedio_debug(save_path,workers,thread)
        else:
            return self.download_vedio_info(save_path,workers,thread)


    def download_vedio_debug(self,save_path=None,workers=1,thread=True):
        '''
        下载视频文件
        save_path: 保存路径
        workers: 任务数目
        thread: True,False采用进程模式
        '''
    
        with tempfile.TemporaryDirectory() as ts_temp_dir:
            logger.debug(f"临时文件夹: {ts_temp_dir}")
            save_path = self.__init_output(save_path)
        
            #多线程下载
            if thread:
                executors = ThreadPoolExecutor(max_workers=workers)
            else:
                executors = ProcessPoolExecutor(max_workers=workers)
            logger.info(f"开始下载，共 ({len(self.url_list)}) 文件")
            tasks = {executors.submit(self._confirm_download_file,file_url,ts_temp_dir):os.path.basename(file_url) for file_url in self.url_list}
            wait(tasks,return_when=ALL_COMPLETED)
            logger.debug("下载完成,开始合并")
            #合并ts文件
            with open(save_path,"wb+") as ww:
                for ts in tasks.values():
                    if os.path.exists(f"{ts_temp_dir}/{ts}"):
                        with open(f"{ts_temp_dir}/{ts}","rb") as rr:
                            ww.write(rr.read())
            logger.success(f"Download {os.path.abspath(save_path)} success")
            return save_path
            


from loguru import logger
import sys

class Log():
    def __init__(self,debug=False):
        if  debug:
            logger.remove()
            logger.add(sys.stdout,level="DEBUG")
            logger.add("em3u8.log",level="DEBUG",rotation="1GB",retention="7 days")
        else:
            logger.remove()
            logger.add(sys.stdout,level="INFO")
            logger.add("em3u8.log",level="INFO",rotation="1GB",retention="7 days")



import re
import requests
from loguru import logger
from urllib.parse import urlparse
from em3u8 import MClient
from Crypto.Cipher import AES
import os


class EM3U8:
    '''
    m3u8 类模块
    基于url或内容创建
    '''


    def __init__(self,content:str = None,url:str = None,timeout=10) -> None:
        self._content = content
        self.url = url
        self.timeout = timeout
        self._tss = []
        self._key = {}

    @property
    def key(self):
        if not self._key:
            key_pattern = re.compile(r'#EXT-X-KEY:([^\n]+)\n')
            if not self._content:
                self._content = self._get_content_from_url()
            keys = key_pattern.findall(self._content)
            if keys:
                self._key = dict(re.findall(r'([^=,]+)=([^,]+)', keys[0]))
            else:
                self._key = {}
        return self._key

    @property
    def content(self):
        if not self._content:
            pass

    @property
    def tss(self):
        '''
        content中获取ts列表
        '''
        if self._tss:
            return self._tss
        if not self._content:
            if not self.url:
                self._tss = []
            else:
                self._content = self._get_content_from_url()
        self._tss = self._get_ts_list_from_content()
        return self._tss
        

    def _get_ts_list_from_content(self):
        '''
        正则获取ts列表
        '''
        if not self._content:
            return []
        ts_pattern = re.compile(r'#EXTINF:[0-9.]+,\n([^\n]+)\n')
        data = ts_pattern.findall(self._content)
        return [self._get_absolute_url(ts_url) for ts_url in data]
            
    def _get_content_from_url(self):
        '''
        从url中获取content,解决302,sub等问题
        '''
        
        url = self.url
        while True:
            logger.debug(f"请求 {url}")
            res = requests.get(url,verify=False,timeout=self.timeout)
            if res.status_code == 302:
                url = res.headers["location"]
                
                logger.debug(f"302 -> {url}")
                continue
            elif res.status_code == 200:
                result,data = self._valid_ts_data(res.text)
                if len(res.text) > 400:
                    logger.debug(res.text[:200] + "\n...\n" + res.text[-200:])
                else:
                    logger.debug(res.text)
                if result:
                    
                    break
                else:
                    url = data
                   
                    logger.debug(f"sub - > {data}")
            else:
                logger.error(f"status code: {res.status_code}")
                logger.error(res.text)
        self.url = url
        self._content = res.text       
        return res.text

    def _valid_ts_data(self,data:str):
        '''
        检测是否是最终m3u8文件
        '''
        if "#EXTINF" in data:
            return True,None
        else:
            for line in data.split("\n"):
                if not line.startswith("#"):
                    new_url = line
                    break
            return False,self._get_baseurl(new_url) + new_url

    def _get_baseurl(self,ts_url):
        '''
        获取baseurl
        '''
        
        if not self.url or ts_url.startswith("http"):
            return ""
        elif ts_url.startswith("/"):
            urlp = urlparse(self.url)
            return urlp.scheme + "://" + urlp.netloc
        else:
            return self.url[0:self.url.rfind("/")+1]
    
    def _get_absolute_url(self,url:str):
        '''
        获取绝对url
        '''
        url = url.replace('"','')
        return self._get_baseurl(url) + url

    def download_vedio(self,save_path=None,workers=1,thread=True,debug=False):
        '''
        下载视频
        '''
        # save_path = MClient(self.tss).download_vedio(save_path,workers,thread)
        save_path = MClient(self.tss).download_vedio(save_path,workers,thread,debug=debug)
        if self.key.get("URI",None):
            logger.debug(f"视频加密，开始解密")
            if os.path.dirname(save_path):
                self.dencrypt(save_path,f"{os.path.dirname(save_path)}/解密-{os.path.basename(save_path)}")
            else:
                self.dencrypt(save_path,f"解密-{save_path}")

    def dencrypt(self,src,dst,key=None,iv=b"\x00"*16):
        '''
        解密
        '''
        if not key:
            key = self.key.get("URI",None)
            key = self._get_absolute_url(key)
        
        if self.key.get("IV",None):
            iv = self.key.get("IV",None)

        
        logger.debug(f"dencrypt: {src} -> {dst}")
        logger.debug(f"dencrypt: key_url -> {key}")
        key = requests.get(key,verify=False).content
        logger.debug(f"dencrypt: key -> {key}-{len(key)},iv:[{iv}]")
        cipher = AES.new(key,AES.MODE_CBC,iv)
        with open(src,"rb") as rr:
            with open(dst,"wb") as ww:
                ww.write(cipher.decrypt(rr.read()))
        logger.debug("dencrypt done")
        
        



import warnings
from urllib3.exceptions import InsecureRequestWarning
from em3u8.httpclient import MClient
from em3u8.m3u8 import EM3U8
from em3u8.parse import Parse
from em3u8.logger import Log

warnings.simplefilter('ignore', InsecureRequestWarning)


from em3u8.sites import Site
import requests
from bs4 import BeautifulSoup
import re
from loguru import logger



class YunBTV(Site):
    '''
    云播
    '''
    name = "云播TV"

    @property
    def m3u8(self):
        if not self.text:
            self.text = requests.get(self.url,verify=False,timeout=20).text
        soup = BeautifulSoup(self.text,"html.parser")
        scripts = soup.find_all("script")
        #js中获取m3u8路径
        for script in scripts:
            if script.get("type") == "text/javascript":
                
                pattern = re.compile(r'"url":"([^"]+)"')
                r = pattern.findall(str(script))
                
                if r:
                    return [r[0].replace("\\","")]
        self.err = "not found m3u8 addr"
        return None

    @property
    def title(self):
        if not self._title:
            if not self.text:
                self.text = requests.get(self.url,verify=False,timeout=20).text
            soup = BeautifulSoup(self.text,"html.parser")
            self._title = "-".join(soup.title.string.split("_")[0:2])
        return self._title



from urllib.parse import urlparse
from em3u8.sites import *
from loguru import logger
import random
import string



class Parse():
    '''
    解析网站类型，获取m3u8地址
    '''
    def __init__(self,url):
        self.__m3u8 = None
        self.__url:str = url
        self.__title = None
        self.__sites = {
            "www.yunbtv.net": YunBTV,
            
        }
        self.__parser = None
        self.err = None
        

    @property
    def title(self):
        try:
            if not self.__title:
                self.__title = self.parser.title
        except Exception as e:
            self.err = e
            self.__title = e
        return self.__title
    
    @property
    def parser(self):
        if not self.__parser:
            # parser_class = self.__sites.get(self.host,Site)
            parser_class = self.__get_parser_class(self.host)
            self.__parser = parser_class(self.__url)
        return self.__parser
    
    def __get_parser_class(self,host):
        '''
        根据host获取解析类
        '''
        for site in self.__sites:
            if site in host:
                return self.__sites[site]
        return Site

    @property
    def m3u8(self):
        try:
            if not self.__m3u8:
                if self.__url.endswith(".m3u8"):
                    self.__m3u8 = self.__url
                    self.__title = f"m3u8_{''.join(random.choices(string.ascii_letters+string.digits,k=10))}"
                else:
                    self.__m3u8 = self.parser.m3u8
                    self.err = self.parser.err
        except Exception as e:
            self.err = e
            self.__m3u8 = None
        
        return self.__m3u8

    @property
    def sites(self):
        return self.__sites

    def show_support(self):
        for url,name in self.__sites.items():
            print("当前版本支持以下站点:")
            print(f"{name.name}:\t\t{url}")

    @property
    def host(self):
        urlp = urlparse(self.__url)
        return urlp.netloc
            


from setuptools import setup

setup(
    name='em3u8',
    version='0.3',
    description='A simple m3u8 package',
    author='xiaotech',
    author_email='xiaotech@163.com',
    packages=['em3u8','em3u8.sites'],
    install_requires=[
        'loguru',
        'requests',
        'beautifulsoup4',
        'tqdm',
        'pycryptodome',
        'click'
    ],
    entry_points={
        'console_scripts': [
            'vget=em3u8.cmd:main'
        ]
    }
)



from bs4 import BeautifulSoup
import requests
import re



class Site():
    name = "SITE"
    
    def __init__(self,url):
        self.url = url
        self.text = None
        self._title = None
        self.err = None
        
        
    @property
    def m3u8(self):
        # 视频资源通用查找
        data = []
        if not self.text:
            self.text = requests.get(self.url,verify=False,timeout=20).text
        soup = BeautifulSoup(self.text,"html.parser")
        scripts = soup.find_all("script")
        #js中获取m3u8路径
        for script in scripts:           
            pattern = re.compile(r"http([^\",\']+)m3u8")
            r = pattern.findall(str(script))            
            if r:
                # data.append(("http"+ r[0] + "m3u8").replace("\\",""))
                data.extend(r)
        #if get m3u8 returned or give err reason
        if data:
            return [("http" + m3u8 + "m3u8").replace("\\","") for m3u8 in data]
        else:
            self.err = "Site: not found m3u8 addr"
            return None

    @property
    def title(self):
        if not self._title:
            if not self.text:
                self.text = requests.get(self.url,verify=False,timeout=20).text
            soup = BeautifulSoup(self.text,"html.parser")
            self._title = soup.title.string
        return self._title



from em3u8 import EM3U8,Parse,Log
from loguru import logger
import click



@click.command
@click.argument("url")
@click.option("--info","-i",help="show resources",is_flag=True)
@click.option("--save-path","-o",help="save vedio path")
@click.option("--workers","-t",help="thread or process counts",default=2,type=int)
@click.option("--process","-p",is_flag=True,help="process model,default thread",default=True)
@click.option("--debug","-d",is_flag=True,help="show debug info")
@click.help_option("--help","-h")
@click.version_option(version="em3u8 v0.2,author: xiaotech@163.com")
def main(url,info,save_path,workers,process,debug):
    
    Log(debug)
    
    site = Parse(url)

    if site.m3u8:
        logger.info(f"名称: {site.title}")
        logger.info(f"资源: {site.m3u8}")
        
        if not info:        
            if not save_path:
                save_path = f"{site.title}.mp4"
            EM3U8(url=site.m3u8[0]).download_vedio(save_path,workers=workers,thread=process,debug=debug)
    else:
        logger.error(f"{site.err}")


if __name__ == "__main__":
    main()